import time
import random
from pyspark_flame import FlameProfiler
from pyspark import SparkConf, SparkContext


def multiply_inefficiently(x):
    for i in range(1000):
        time.sleep(0.0001 * random.random())
        time.sleep(0.0001 * random.random())
    return x * 2


conf = SparkConf().set("spark.python.profile", "true")#.set("spark.python.profile.dump", ".")
sc = SparkContext('local', 'test', conf=conf, profiler_cls=FlameProfiler, environment={'pyspark_flame.interval': 0.25})
sc.parallelize(range(1000)).map(multiply_inefficiently).take(10)
sc.show_profiles()
sc.dump_profiles('.')
sc.stop()



from pyspark_flame import FlameProfiler
from pyspark import SparkConf, SparkContext
import os
import shutil
import time
import tempfile
import unittest


def wait_a_bit(*args, **kwargs):
    time.sleep(5.0)
    return 1


class PysparkFlameTest(unittest.TestCase):
    def setUp(self):
        self.dumpdir = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.dumpdir)

    def test_pyspark_flame(self):
        conf = SparkConf().set("spark.python.profile", "true")
        sc = SparkContext('local[*]', 'test',
                          conf=conf,
                          profiler_cls=FlameProfiler,
                          environment={'pyspark_flame.interval': 0.25})
        sc.parallelize(range(4)).map(wait_a_bit).sum()
        sc.dump_profiles(self.dumpdir)
        dumps = os.listdir(self.dumpdir)
        self.assertEqual(1, len(dumps))
        with open(os.path.join(self.dumpdir, dumps[0])) as dumpfile:
            for line in dumpfile.readlines():
                location, count = line.split(' ')
                if 'pyspark_flame_test.py:wait_a_bit:11' in location:
                    count = int(count)
                    self.assertIn(count, range(70, 90))
                    return
            else:
                self.fail('No wait_a_bit profile line found')



from collections import defaultdict
from pyspark import AccumulatorParam
from pyspark.profiler import Profiler
from os.path import join
from threading import Thread, Event
from sys import _current_frames


def extract_stack(frame):
    result = []
    while frame is not None:
        result.append((
            frame.f_code.co_filename,
            frame.f_code.co_name,
            frame.f_lineno
        ))
        frame = frame.f_back
    return tuple(reversed(result))


class Collector(object):
    def __init__(self, interval=0.01):
        self.interval = interval
        self.finished = Event()

    def start(self):
        self.thread = Thread(target=self.run)
        self.thread.start()

    def stop(self):
        self.finished.set()
        self.thread.join()

    def run(self):
        results = defaultdict(int)
        self.results = results
        this_thread = self.thread.ident
        while True:
            self.finished.wait(self.interval)
            if self.finished.is_set():
                return
            results = self.results
            frames = _current_frames()
            for frame_id, frame in frames.items():
                if frame_id != this_thread:
                    stack = extract_stack(frame)
                    results[stack] += 1


class ResultsAccumulator(AccumulatorParam):
    def zero(self, value):
        return defaultdict(int)

    def addInPlace(self, a, b):
        result = defaultdict(int)
        for r in (a, b):
            for stack, count in r.items():
                result[stack] += count
        return result


class FlameProfiler(Profiler):
    def __init__(self, ctx):
        self.interval = float(ctx.environment.get('pyspark_flame.interval', 0.05))
        self._accumulator = ctx.accumulator(defaultdict(int), ResultsAccumulator())

    def profile(self, func):
        collector = Collector(self.interval)
        collector.start()
        try:
            func()
        finally:
            collector.stop()
            return self._accumulator.add(collector.results)

    def stats(self):
        return self._accumulator.value

    def show(self, id):
        print "Flame Data for RDD {}".format(id)
        print self.format()

    def dump(self, id, path):
        with open(join(path, 'rdd-{}.flame'.format(id)), 'w') as f:
            f.write(self.format())

    def format(self):
        return ''.join(
            '{stack} {count}\n'.format(stack=';'.join(
                '{file}:{method}:{line}'.format(file=file_, method=method, line=line)
                for file_, method, line in stack
            ), count=count)
            for stack, count in self.stats().items()
        )



#!/usr/bin/env python

from setuptools import setup, find_packages

setup(
    name='pyspark-flame',
    version='0.2.0',
    description='A low-overhead sampling profiler for PySpark, that outputs Flame Graphs',
    author='James Pickering',
    author_email='james_pic@hotmail.com',
    license='MIT',
    url='https://github.com/jamespic/pyspark-flame',
    download_url='https://github.com/jamespic/pyspark-flame/archive/0.1.2.tar.gz',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    scripts=['FlameGraph/flamegraph.pl'],
    test_suite='test'
)
